(* type input *)
type input = Strng.t
(* type result *)
type 'a result = ('a * input)option
(* type t *)
type 'a t = input -> 'a result
(* test : 'a t -> string -> 'a *)
let test p s = match p (Strng.of_string s) with
 | None      -> failwith "no parse possible"
 | Some(x,_) -> x
(* parse : 'a t -> input -> 'a option *)
let parse p s = match p (Strng.of_string s) with
 | None      -> None
 | Some(x,_) -> Some x
(* return : 'a -> 'a t *)
let return x = fun i -> Some(x,i)
(* bind : 'a t -> ('a -> 'b t) -> 'b t *)
let (>>=) p f i = match p i with None      -> None
                               | Some(x,i) -> f x i
(* choice : 'a t -> 'a t -> 'a t *)
let (<|>) p q i = match p i with None        -> q i
                               | Some _ as r -> r
(* then : 'a t -> 'b t -> 'b t *)
let (>>) p q = p >>= (fun _ -> q)
(* many : 'a t -> ('a list)t *)
let rec many p =
 (p >>= fun x -> many p >>= fun xs -> return(x::xs))
 <|> return []
(* many1 : 'a t -> ('a list)t *)
let many1 p = p >>= fun x -> many p >>= fun xs -> return(x::xs)
(* between : 'a t -> 'b t -> 'c t -> 'b t *)
let between l p r = l >> p >>= fun x -> r >> return x
(* sep_by1 : 'a t -> 'b t -> ('b list)t *)
let sep_by1 sep p = p >>= fun x -> many (sep >> p) >>= fun xs -> return(x::xs)
(* sep_by : 'a t -> 'b t -> ('b list)t *)
let sep_by sep p = sep_by1 sep p <|> return []
(* eoi : unit t *)
let eoi = function [] -> Some((),[])
                 | _  -> None
(* sat : (char -> bool) -> char t *)
let sat p = function []    -> None
                   | c::cs -> if p c then Some(c,cs)
                                     else None
(* any : char t *)
let any = sat (fun _ -> true)
(* char : char -> char t *)
let char c = sat ((=)c)
(* letter : char t *)
let letter = sat (fun c -> ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z'))
(* digit : char t *)
let digit = sat (fun c -> '0' <= c && c <= '9')
(* oneof : string -> char t *)
let oneof s = sat (fun c -> Lst.mem c (Strng.of_string s))
(* noneof : string -> char t *)
let noneof s = sat (fun c -> not(Lst.mem c (Strng.of_string s)))
(* space : char t *)
let space = oneof " \t\n"
(* spaces : unit t *)
let spaces = many space >> return ()
(* string : string -> Strng.t t *)
let string s cs =
 let ss = Strng.of_string s in
 let rec string = function ([],cs)       -> Some(ss,cs)
                         | (_,[])        -> None
                         | (s::ss,c::xs) -> if s = c then string(ss,xs)
                                                     else None
 in
 string(ss,cs)
(* *)
